/* 
 *  Vnc2swf - writeswf.c (generates a SWF file with Ming)
 *
 *  Copyright (C) 2002 by Yusuke Shinyama (yusuke at cs . nyu . edu)
 *  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <vncviewer.h>
#include <sys/time.h>
#include <math.h>

/* ming */
#include <ming.h>
#include <zlib.h>

#define DEBUG_SWF 0
#define DEBUG_DBL 0
#define DEBUG_ZBUFFER 0

#define CHANNELS 4

static FILE* dumpf;
static int movie_width, movie_height;
static SWFMovie movie;
static unsigned long tmpbuf_len;
static char* tmpbuf_src;
static char* tmpbuf_compressed;
static long intervalusec;
static struct timeval tv0;
static int frame_rate, nframes;
static int macromedia_hack;
static unsigned long expanded_image_bytes;


/* ZBuffer: covered pixels = 1 */
typedef unsigned short zpixel;
typedef struct _DisplayItem {
  SWFDisplayItem d;
  zpixel id;
  int x, y, w, h;
  struct _DisplayItem* next;
  struct _DisplayItem* prev;
} DisplayItem;

static zpixel shapeid;
static zpixel* zbuffer;
static DisplayItem items;


/* initializeZBuffer:
 */
static Bool initializeZBuffer(void)
{
  zpixel* p;
  unsigned long buflen;
  buflen = (unsigned long)movie_width*(unsigned long)movie_height*sizeof(zpixel);
  p = zbuffer = (zpixel*)malloc(buflen);
  if (!zbuffer) {
    perror("initializeZBuffer: Out of memory");
    return False;
  }
  memset(zbuffer, 0, buflen);
  shapeid = 1;
  items.d = NULL;
  items.next = &items;
  items.prev = &items;
  return True;
}


/* paintZBuffer:
 *   Fill the given rect with the current shape id.
 */
static Bool paintZBuffer(SWFDisplayItem d, int x, int y, int w, int h)
{
  zpixel* p = zbuffer + y * movie_width + x;
  DisplayItem* di;

#if DEBUG_ZBUFFER
  fprintf(stderr,"paintZBuffer: %u: (%d,%d)-(%d,%d)\n",
	  shapeid, x, y, x+w, y+h);
#endif
  di = (DisplayItem*)malloc(sizeof(DisplayItem));
  if (!di) {
    perror("fillZBuffer: Out of memory");
    return False;
  }
  di->d = d;
  di->id = shapeid;
  di->x = x;
  di->y = y;
  di->w = w;
  di->h = h;
  /* WARNING: h is broken after this! */
  while(h--) {
    zpixel* p1 = p;
    int w1 = w;
    while(w1--) {
      *(p1++) = shapeid;
    }
    p += movie_width;
  }

  /* add to the list */
  di->next = items.next;
  di->prev = &items;
  items.next->prev = di;
  items.next = di;
  shapeid++;
  return True;
}


/* checkZBufferCovered:
 *   Check if every pixel in the given rect is covered thereafter.
 */
static Bool checkZBufferCovered(zpixel id, int x, int y, int w, int h)
{
  zpixel* p = zbuffer + y * movie_width + x;
  /* WARNING: h is broken after this! */
  while(h--) {
    zpixel* p1 = p;
    int w1 = w;
    while(w1--) {
      if (*(p1++) <= id) return False; /* this pixel is not yet covered! */
    }
    p += movie_width; 
  }
  return True;
}


/* checkRemoveItems:
 *   Remove covered items from the movie.
 */
static void checkRemoveItems(void)
{
  DisplayItem* di = items.next;

  while(di->d) {
    if (checkZBufferCovered(di->id, di->x, di->y, di->w, di->h)) {
      DisplayItem* removed = di;
#if DEBUG_ZBUFFER
      fprintf(stderr,"checkRemoveItems: covered: %u: (%d,%d)-(%d,%d)\n",
	      di->id, di->x, di->y, di->x+di->w, di->y+di->h);
#endif
      SWFMovie_remove(movie, di->d);

      /* remove from the list */
      di->next->prev = di->prev;
      di->prev->next = di->next;
      di = di->next;
      free((char*)removed);
    } else {
      di = di->next;
    }
  }
}


/* checkNextFrame:
 *   Add frames necessary.
 */
static void checkNextFrame(void)
{
  struct timeval tv;
  struct timezone dummytz;
  gettimeofday(&tv, &dummytz);
  if (tv0.tv_sec < 0) {		/* initialize */
    tv0 = tv;
    tv0.tv_usec += intervalusec;
    if (1000000L <= tv0.tv_usec) {
      tv0.tv_usec -= 1000000L;
      tv0.tv_sec++;
    }
  }
  while(tv0.tv_sec < tv.tv_sec || 
	(tv0.tv_sec == tv.tv_sec && tv0.tv_usec < tv.tv_usec)) {
    SWFMovie_nextFrame(movie);
    nframes++;
#if DEBUG_SWF
    fprintf(stderr, "checkNextFrame: frame: %d\n", nframes);
#endif /* DEBUG_SWF */
    tv0.tv_usec += intervalusec;
    if (1000000L <= tv0.tv_usec) {
      tv0.tv_usec -= 1000000L;
      tv0.tv_sec++;
    }    
  }
}


/* ConvBufToRGB:
 *   Convert a framebuffer image (ZPixmap) to a RGB array.
 */
static void ConvBufToRGB(unsigned char* buf, int w, int h, int skip)
{
  int i, j;
  char* p = tmpbuf_src;

  switch(myFormat.bitsPerPixel) {
  case 8:
    break;
  case 16:
    skip *= 2;
    break;
  case 32:
    skip *= 4;
    break;
  }

  for(i = 0; i < h; i++) {
    unsigned char* linebuf = buf;
    for(j = 0; j < w; j++) {
      unsigned long c, c1, c2, c3;
      switch(myFormat.bitsPerPixel) {
      case 8:
	if (appData.useBGR233) {
	  c = BGR233ToPixel[*(linebuf++)];
	} else {
	  c = *(linebuf++);
	}
	break;
      case 16:
	c1 = *(linebuf++);
	c = (*(linebuf++) << 8) | c1;
	break;
      case 32:
	c1 = *(linebuf++);
	c2 = *(linebuf++);
	c3 = *(linebuf++);
	c = (*(linebuf++) << 24) | (c3 << 16) | (c2 << 8) | c1;
	break;
      }
      /* RGB */
      *(p++) = 0;
      *(p++) = redRv[(c>>myFormat.redShift) & myFormat.redMax]; /* R */
      *(p++) = greenRv[(c>>myFormat.greenShift) & myFormat.greenMax]; /* G */
      *(p++) = blueRv[(c>>myFormat.blueShift) & myFormat.blueMax]; /* B */
    }
    buf += skip;
  }
}


/* ImageToDBL:
 *   Make a DBL image from raw framebuffers.
 *
 * DBL format:
 *   3bytes: 'DBl' (header)
 *   1byte : format (1:noalpha, 2:hasalpha)
 *   4bytes: size of the data contained below (n+6 or n+5)
 *   1byte : palette? (5:nopalette, 3:haspalette)
 *   2bytes: width
 *   2bytes: height
 *   [1byte: num_palette-1] (if has a palette)
 *   nbytes: compressed RGB or RGBA
 */
/* write_ulong */
static char* write_ulong(char* p, unsigned long x)
{
  *(p++) = (x >> 24) & 0xff;
  *(p++) = (x >> 16) & 0xff;
  *(p++) = (x >> 8) & 0xff;
  *(p++) = (x >> 0) & 0xff;
  return p;
}
/* write_ushort */
static char* write_ushort(char* p, unsigned short x)
{
  *(p++) = (x >> 0) & 0xff;
  *(p++) = (x >> 8) & 0xff;
  return p;
}
static char* ImageToDBL(char* buf, int w, int h, int skip, unsigned long* len)
{
  char* dblbuf;
  char* p;
  unsigned long dbllen, dbldatalen, dbltotallen;
  unsigned long imglen = (unsigned long)w * (unsigned long)h * CHANNELS;

  ConvBufToRGB(buf, w, h, skip);
  dbllen = tmpbuf_len;
  compress(tmpbuf_compressed, &dbllen, tmpbuf_src, imglen);
  dbldatalen = dbllen+1+2+2;
  dbltotallen = dbldatalen+3+1+4;
  dblbuf = malloc(dbltotallen);
  if (!dblbuf) return(NULL);
  expanded_image_bytes += imglen;
  
  p = dblbuf;
  /* a DBL header */
  *(p++) = 'D';
  *(p++) = 'B';
  *(p++) = 'l';
  /* no alpha */
  *(p++) = 1;
  /* the dummy data size */
  p = write_ulong(p, dbldatalen);
  /* no pallete */
  *(p++) = 5;			
  /* the image size */
  p = write_ushort(p, w);
  p = write_ushort(p, h);
  /* the compressed data */
  memcpy(p, tmpbuf_compressed, dbllen);

  *len = dbltotallen;

#if DEBUG_DBL
  {
    /* dump each DBL file into files 
     * CAUTION: files can be too big and many! 
     */
    char fname[256];
    FILE* fp; 
    sprintf(fname, "f%05u.dbl", shapeid);
    fp = fopen(fname, "wb");
    fwrite(dblbuf, 1, dbltotallen, fp);
    fclose(fp);
  }
#endif  

  return(dblbuf);
}


/* WriteInitMovie:
 *   Initialize Ming and prepare a movie file.
 */
Bool WriteInitMovie(char* filename, int frate, int mmhack, int w, int h)
{
  fprintf(stderr,"WriteInitMovie: Opening file: \"%s\" with frame rate %d\n", 
	  filename, frate);
  dumpf = fopen(filename, "wb");
  if (!dumpf) {
    perror("WriteInitMovie: Cannot open file");
    return False;
  }
  
  movie_width = w;
  movie_height = h;
  tmpbuf_len = (unsigned long)floor((unsigned long)w*(unsigned long)h*CHANNELS*1.01+12);
  tmpbuf_src = malloc(tmpbuf_len);
  tmpbuf_compressed = malloc(tmpbuf_len);
  if (!tmpbuf_src || !tmpbuf_compressed) {
    perror("WriteInitMovie: Out of memory");
    return False;
  }
  initializeZBuffer();
  
  movie = newSWFMovie();
  Ming_init();
  SWFMovie_setRate(movie, frate);
  SWFMovie_setDimension(movie, w, h);
  intervalusec = 1000000L / frate;
  tv0.tv_sec = tv0.tv_usec = -1;
  nframes = 0;
  frame_rate = frate;
  macromedia_hack = mmhack;
  return True;
}


/* WriteFinishMovie:
 *   Write the movie file.
 */
Bool WriteFinishMovie(void)
{
  long len;

  /* flush */
  checkNextFrame();

  fprintf(stderr, "WriteFinishMovie: %d frames (%.1f sec), expanded_image: %lu\n",
	  nframes, (double)nframes/(double)frame_rate, expanded_image_bytes);
  
  len = SWFMovie_output(movie, fileOutputMethod, dumpf);
  destroySWFMovie(movie);

  if (0 < len) {
    fprintf(stderr, "WriteFinishMovie: %ld bytes dumped.\n", len);
  } else {
    perror("WriteFinishMovie: cannot write the file!");
    return False;
  }
  return True;
}


/* WriteRawImage:
 *   Put an image obtained from the framebuffer.
 */
Bool WriteRawImage(char* buf, int x, int y, int w, int h, int skip)
{
  SWFShape shape;
  SWFFill fill;
  SWFDBLBitmap dbl;
  SWFDisplayItem d;
  SWFInput input;
  char* dblbuf;
  unsigned long dbllen;

#if DEBUG_SWF
  fprintf(stderr,"WriteRawImage: %dx%d to (%d, %d)\n", w, h, x, y);
#endif /* DEBUG_SWF */

  checkNextFrame();
  dblbuf = ImageToDBL(buf, w, h, skip, &dbllen);
  if (!dblbuf) return(False);
  input = newSWFInput_allocedBuffer(dblbuf, dbllen);
  dbl = newSWFDBLBitmap_fromInput(input);
  shape = newSWFShape();
  fill = SWFShape_addBitmapFill(shape, dbl, SWFFILL_TILED_BITMAP);
  SWFShape_setRightFill(shape, fill);
  if (macromedia_hack) {
    SWFShape_movePenTo(shape, 1, 1);
  } else {
    SWFShape_movePenTo(shape, 0, 0);
  }
  SWFShape_drawLine(shape, w, 0);
  SWFShape_drawLine(shape, 0, h);
  SWFShape_drawLine(shape,-w, 0);
  SWFShape_drawLine(shape, 0,-h);
  SWFShape_end(shape);
  d = SWFMovie_add(movie, shape);
  SWFDisplayItem_moveTo(d, x, y);

  paintZBuffer(d, x, y, w, h);
  checkRemoveItems();
  return True;
}


/* WriteMoveImage:
 *   Scroll the image.
 */
Bool WriteMoveImage(Display* dpy, Drawable d, XImage* img,
		    int x0, int y0, int w, int h, int x1, int y1)
{
  int scrWidthInBytes = si.framebufferWidth * myFormat.bitsPerPixel / 8;
  char* buf = image->data + y0 * scrWidthInBytes + x0 * myFormat.bitsPerPixel / 8;
  XGetSubImage(dpy, d, x0, y0, w, h, 0xffffffff, ZPixmap, image, x0, y0);
  WriteRawImage(buf, x1, y1, w, h, si.framebufferWidth);
  return True;
}


/* WriteRectangle:
 *   Put a rectangle.
 */
Bool WriteRectangle(Pixel c, int x, int y, int w, int h)
{
  SWFShape shape;
  SWFFill fill;
  SWFDisplayItem d;
  int red, green, blue;

#if DEBUG_SWF
  fprintf(stderr,"WriteRectanble: %dx%d to (%d, %d)\n", w, h, x, y);
#endif /* DEBUG_SWF */

  shape = newSWFShape();
  checkNextFrame();
  red = redRv[(c>>myFormat.redShift) & myFormat.redMax];
  green = greenRv[(c>>myFormat.greenShift) & myFormat.greenMax];
  blue = blueRv[(c>>myFormat.blueShift) & myFormat.blueMax];
  fill = SWFShape_addSolidFill(shape, red, green, blue, 255);
  SWFShape_setRightFill(shape, fill);
  if (macromedia_hack) {
    SWFShape_movePenTo(shape, 1, 1);
  } else {
    SWFShape_movePenTo(shape, 0, 0);
  }
  SWFShape_drawLine(shape, w, 0);
  SWFShape_drawLine(shape, 0, h);
  SWFShape_drawLine(shape, -w, 0);
  SWFShape_drawLine(shape, 0, -h);
  d = SWFMovie_add(movie, shape);
  SWFDisplayItem_moveTo(d, x, y);

  paintZBuffer(d, x, y, w, h);
  checkRemoveItems();
  return True;
}
